#pylint: disable=no-name-in-module,unused-import
import numpy as np
from .factory import get_planned_FFT
from .utilities import FFTW_FORWARD, FFTW_BACKWARD, FFTW_REDFT00, FFTW_REDFT01, \
    FFTW_REDFT10, FFTW_REDFT11, FFTW_RODFT00, FFTW_RODFT01, FFTW_RODFT10, \
    FFTW_RODFT11, FFTW_MEASURE, FFTW_DESTROY_INPUT, FFTW_UNALIGNED, \
    FFTW_CONSERVE_MEMORY, FFTW_EXHAUSTIVE, FFTW_PRESERVE_INPUT, FFTW_PATIENT, \
    FFTW_ESTIMATE, FFTW_WISDOM_ONLY, C2C_FORWARD, C2C_BACKWARD, R2C, C2R, \
    FFTW_R2HC, FFTW_HC2R, FFTW_DHT, get_alignment, aligned, aligned_like

flag_dict = {key: val for key, val in locals().items()
             if key.startswith('FFTW_')}

dct_type = {
    1: FFTW_REDFT00,
    2: FFTW_REDFT10,
    3: FFTW_REDFT01,
    4: FFTW_REDFT11}

idct_type = {
    1: FFTW_REDFT00,
    2: FFTW_REDFT01,
    3: FFTW_REDFT10,
    4: FFTW_REDFT11}

dst_type = {
    1: FFTW_RODFT00,
    2: FFTW_RODFT10,
    3: FFTW_RODFT01,
    4: FFTW_RODFT11}

idst_type = {
    1: FFTW_RODFT00,
    2: FFTW_RODFT01,
    3: FFTW_RODFT10,
    4: FFTW_RODFT11}

def fftn(input_array, s=None, axes=(-1,), threads=1,
         flags=(FFTW_MEASURE,), output_array=None):
    """Return complex-to-complex forward transform object

    Parameters
    ----------
    input_array : complex array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the FFT.
    threads : int, optional
        Number of threads used in computing FFT.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : complex array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for complex-to-complex transforms

    Note
    ----
    This routine does not compute the fftn, it merely returns an instance of
    a class that can do it.
    The contents of the input_array may be overwritten during planning. Make
    sure to keep a copy if needed.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import fftn as plan_fftn
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='D')
    >>> fftn = plan_fftn(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = fftn()
    >>> print(B)
    [10.+0.j -2.+2.j -2.+0.j -2.-2.j]
    >>> assert id(A) == id(fftn.input_array)
    >>> assert id(B) == id(fftn.output_array)

    """
    kind = FFTW_FORWARD
    assert input_array.dtype.char in 'FDG'
    if output_array is None:
        n = get_alignment(input_array)
        output_array = aligned(input_array.shape, n,
                               input_array.dtype.char.upper())
    else:
        assert input_array.shape == output_array.shape
        assert output_array.dtype.char == input_array.dtype.char.upper()
    M = np.prod(np.take(input_array.shape, axes))
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, 1.0/M)

def ifftn(input_array, s=None, axes=(-1,), threads=1,
          flags=(FFTW_MEASURE,), output_array=None):
    """
    Return complex-to-complex inverse transform object

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the inverse FFT.
    threads : int, optional
        Number of threads used in computing FFT.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for complex-to-complex inverse
        transforms

    Note
    ----
    This routine does not compute the ifftn, it merely returns an instance of
    a class that can do it.
    The contents of the input_array may be overwritten during planning. Make
    sure that you keep a copy if needed.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import ifftn as plan_ifftn
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, FFTW_PRESERVE_INPUT, aligned
    >>> A = aligned(4, dtype='D')
    >>> ifftn = plan_ifftn(A, flags=(FFTW_ESTIMATE, FFTW_PRESERVE_INPUT))
    >>> A[:] = 1, 2, 3, 4
    >>> B = ifftn()
    >>> print(B)
    [10.+0.j -2.-2.j -2.+0.j -2.+2.j]
    >>> assert id(B) == id(ifftn.output_array)
    >>> assert id(A) == id(ifftn.input_array)

    """
    kind = FFTW_BACKWARD
    assert input_array.dtype.char in 'FDG'
    if output_array is None:
        output_array = aligned_like(input_array)
    else:
        assert input_array.shape == output_array.shape
    M = np.prod(np.take(input_array.shape, axes))
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, 1.0/M)

def rfftn(input_array, s=None, axes=(-1,), threads=1,
          flags=(FFTW_MEASURE,), output_array=None):
    """Return real-to-complex transform object

    Parameters
    ----------
    input_array : real array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real to complex FFT.
    threads : int, optional
        Number of threads used in computing FFT.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for real-to-complex transforms

    Note
    ----
    This routine does not compute the rfftn, it merely returns an instance of
    a class that can do it.
    The contents of the input_array may be overwritten during planning. Make
    sure that you keep a copy if needed.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import rfftn as plan_rfftn
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='d')
    >>> rfftn = plan_rfftn(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = rfftn()
    >>> print(B)
    [10.+0.j -2.+2.j -2.+0.j]
    >>> assert id(A) == id(rfftn.input_array)
    >>> assert id(B) == id(rfftn.output_array)

    """
    kind = R2C
    assert input_array.dtype.char in 'fdg'
    if output_array is None:
        sz = list(input_array.shape)
        sz[axes[-1]] = input_array.shape[axes[-1]]//2+1
        dtype = input_array.dtype.char
        n = get_alignment(input_array)
        output_array = aligned(sz, n=n, dtype=np.dtype(dtype.upper()))
    else:
        assert input_array.shape[axes[-1]]//2+1 == output_array.shape[axes[-1]]
    M = np.prod(np.take(input_array.shape, axes))
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, 1.0/M)

def irfftn(input_array, s=None, axes=(-1,), threads=1,
           flags=(FFTW_MEASURE,), output_array=None):
    """Return inverse complex-to-real transform object

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Shape of output array along each of the transformed axes. Must be same
        length as axes (len(s) == len(axes)). If not given it is assumed that
        the shape of the output along the first transformed axis (i.e.,
        axes[-1]) is an even number. It is not possible to determine exactly,
        because for a real transform the output of a real array of length N is
        N//2+1. However, both N=4 and N=5 gives 4//2+1=3 and 5//2+1=3, so it is
        not possible to determine whether 4 or 5 is correct. Hence it must be
        given.
    axes : sequence of ints, optional
        Axes over which to compute the real to complex FFT.
    threads : int, optional
        Number of threads used in computing FFT.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for complex-to-real transforms

    Note
    ----
    This routine does not compute the irfftn, it merely returns an instance of
    a class that can do it.
    The irfftn is not possible to use with the FFTW_PRESERVE_INPUT flag.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import irfftn as plan_irfftn
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='D')
    >>> irfftn = plan_irfftn(A, flags=(FFTW_ESTIMATE,)) # no shape given for output
    >>> A[:] = 1, 2, 3, 4
    >>> B = irfftn()
    >>> print(B)
    [15. -4.  0. -1.  0. -4.]
    >>> irfftn = plan_irfftn(A, s=(7,), flags=(FFTW_ESTIMATE,)) # output shape given
    >>> B = irfftn()
    >>> print(B)
    [19.         -5.04891734 -0.30797853 -0.64310413 -0.64310413 -0.30797853
     -5.04891734]
    >>> assert id(B) == id(irfftn.output_array)
    >>> assert id(A) == id(irfftn.input_array)

    """
    kind = C2R
    assert input_array.dtype.char in 'FDG'
    assert FFTW_PRESERVE_INPUT not in flags
    sz = list(input_array.shape)
    if s is not None:
        assert len(axes) == len(s)
        for q, axis in zip(s, axes):
            sz[axis] = q
    else:
        sz[axes[-1]] = 2*sz[axes[-1]]-2
    if output_array is None:
        dtype = input_array.dtype.char
        n = get_alignment(input_array)
        output_array = aligned(sz, n=n, dtype=np.dtype(dtype.lower()))
    else:
        assert list(output_array.shape) == sz

    assert sz[axes[-1]]//2+1 == input_array.shape[axes[-1]]
    M = np.prod(np.take(output_array.shape, axes))
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, 1.0/M)

def dctn(input_array, s=None, axes=(-1,), type=2, threads=1,
         flags=(FFTW_MEASURE,), output_array=None):
    """Return discrete cosine transform object

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real-to-real dct.
    type : int, optional
        Type of `dct <http://www.fftw.org/fftw3_doc/Real_002dto_002dReal-Transform-Kinds.html>`_

            - 1 - FFTW_REDFT00
            - 2 - FFTW_REDFT10,
            - 3 - FFTW_REDFT01,
            - 4 - FFTW_REDFT11
    threads : int, optional
        Number of threads used in computing dct.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for real-to-real dct transforms
        of given type

    Note
    ----
    This routine does not compute the dct, it merely returns an instance of
    a class that can do it.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import dctn as plan_dct
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='d')
    >>> dct = plan_dct(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = dct()
    >>> print(B)
    [20.         -6.30864406  0.         -0.44834153]
    >>> assert id(A) == id(dct.input_array)
    >>> assert id(B) == id(dct.output_array)

    """
    assert input_array.dtype.char in 'fdg'
    if output_array is None:
        output_array = aligned_like(input_array)
    else:
        assert input_array.shape == output_array.shape
    kind = dct_type[type]
    kind = [kind]*len(axes)
    M = get_normalization(kind, input_array.shape, axes)
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, M)

def idctn(input_array, s=None, axes=(-1,), type=2, threads=1,
          flags=(FFTW_MEASURE,), output_array=None):
    """Return inverse discrete cosine transform object

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real-to-real idct.
    type : int, optional
        Type of `idct <http://www.fftw.org/fftw3_doc/Real_002dto_002dReal-Transform-Kinds.html>`_

            - 1 - FFTW_REDFT00
            - 2 - FFTW_REDFT01
            - 3 - FFTW_REDFT10
            - 4 - FFTW_REDFT11
    threads : int, optional
        Number of threads used in computing idct.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for real-to-real idct transforms
        of given type

    Note
    ----
    This routine does not compute the idct, it merely returns an instance of
    a class that can do it.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import idctn as plan_idct
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='d')
    >>> idct = plan_idct(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = idct()
    >>> print(B)
    [11.99962628 -9.10294322  2.61766184 -1.5143449 ]
    >>> assert id(A) == id(idct.input_array)
    >>> assert id(B) == id(idct.output_array)

    """
    assert input_array.dtype.char in 'fdg'
    if output_array is None:
        output_array = aligned_like(input_array)
    else:
        assert input_array.shape == output_array.shape
    kind = idct_type[type]
    kind = [kind]*len(axes)
    M = get_normalization(kind, input_array.shape, axes)
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, M)

def dstn(input_array, s=None, axes=(-1,), type=2, threads=1,
         flags=(FFTW_MEASURE,), output_array=None):
    """Return discrete sine transform object

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real-to-real dst.
    type : int, optional
        Type of `dst <http://www.fftw.org/fftw3_doc/Real_002dto_002dReal-Transform-Kinds.html>`_

            - 1 - FFTW_RODFT00
            - 2 - FFTW_RODFT10
            - 3 - FFTW_RODFT01
            - 4 - FFTW_RODFT11
    threads : int, optional
        Number of threads used in computing dst.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for real-to-real dst transforms
        of given type

    Note
    ----
    This routine does not compute the dst, it merely returns an instance of
    a class that can do it.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import dstn as plan_dst
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='d')
    >>> dst = plan_dst(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = dst()
    >>> print(B)
    [13.06562965 -5.65685425  5.411961   -4.        ]
    >>> assert id(A) == id(dst.input_array)
    >>> assert id(B) == id(dst.output_array)

    """
    assert input_array.dtype.char in 'fdg'
    if output_array is None:
        output_array = aligned_like(input_array)
    else:
        assert input_array.shape == output_array.shape
    kind = dst_type[type]
    kind = [kind]*len(axes)
    M = get_normalization(kind, input_array.shape, axes)
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, M)

def idstn(input_array, s=None, axes=(-1,), type=2, threads=1,
          flags=(FFTW_MEASURE,), output_array=None):
    """Return inverse discrete sine transform object

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real-to-real inverse dst.
    type : int, optional
        Type of `idst <http://www.fftw.org/fftw3_doc/Real_002dto_002dReal-Transform-Kinds.html>`_

            - 1 - FFTW_RODFT00
            - 2 - FFTW_RODFT01
            - 3 - FFTW_RODFT10
            - 4 - FFTW_RODFT11
    threads : int, optional
        Number of threads used in computing inverse dst.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for real-to-real idst transforms
        of given type

    Note
    ----
    This routine does not compute the idst, it merely returns an instance of
    a class that can do it.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import idstn as plan_idst
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='d')
    >>> idst = plan_idst(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = idst()
    >>> print(B)
    [13.13707118 -1.6199144   0.72323135 -0.51978306]
    >>> assert id(A) == id(idst.input_array)
    >>> assert id(B) == id(idst.output_array)

    """
    assert input_array.dtype.char in 'fdg'
    if output_array is None:
        output_array = aligned_like(input_array)
    else:
        assert input_array.shape == output_array.shape
    kind = idst_type[type]
    kind = [kind]*len(axes)
    M = get_normalization(kind, input_array.shape, axes)
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, M)

def ihfftn(input_array, s=None, axes=(-1,), threads=1,
           flags=(FFTW_MEASURE,), output_array=None):
    """Return inverse transform object for an array with Hermitian symmetry

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the ihfftn.
    threads : int, optional
        Number of threads used in computing ihfftn.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for real-to-complex ihfftn
        transforms

    Note
    ----
    This routine does not compute the ihfttn, it merely returns an instance of
    a class that can do it.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import ihfftn as plan_ihfftn
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='d')
    >>> ihfftn = plan_ihfftn(A, flags=(FFTW_ESTIMATE,))
    >>> A[:] = 1, 2, 3, 4
    >>> B = ihfftn()
    >>> print(B)
    [10.+0.j -2.+2.j -2.+0.j]
    >>> assert id(A) == id(ihfftn.input_array)
    >>> assert id(B) == id(ihfftn.output_array)

    """
    kind = R2C
    assert input_array.dtype.char in 'fdg'
    if output_array is None:
        dtype = input_array.dtype.char
        sz = list(input_array.shape)
        sz[axes[-1]] = input_array.shape[axes[-1]]//2+1
        n = get_alignment(input_array)
        output_array = aligned(sz, n=n, dtype=np.dtype(dtype.upper()))
    else:
        assert input_array.shape[axes[-1]]//2+1 == output_array.shape[axes[-1]]
    M = get_normalization(kind, input_array.shape, axes)
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, M)

def hfftn(input_array, s=None, axes=(-1,), threads=1,
          flags=(FFTW_MEASURE,), output_array=None):
    """Return transform object for an array with Hermitian symmetry

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the hfftn.
    threads : int, optional
        Number of threads used in computing hfftn.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Returns
    -------
    :class:`.fftwf_xfftn.FFT`, :class:`.fftw_xfftn.FFT` or :class:`.fftwl_xfftn.FFT`
        An instance of the return type configured for complex-to-real hfftn
        transforms

    Note
    ----
    This routine does not compute the hfttn, it merely returns an instance of
    a class that can do it.

    Examples
    --------
    >>> import numpy as np
    >>> from mpi4py_fft.fftw import hfftn as plan_hfftn
    >>> from mpi4py_fft.fftw import FFTW_ESTIMATE, aligned
    >>> A = aligned(4, dtype='D')
    >>> hfftn = plan_hfftn(A, flags=(FFTW_ESTIMATE,)) # no shape given for output
    >>> A[:] = 1, 2, 3, 4
    >>> B = hfftn()
    >>> print(B)
    [15. -4.  0. -1.  0. -4.]
    >>> hfftn = plan_hfftn(A, s=(7,), flags=(FFTW_ESTIMATE,)) # output shape given
    >>> B = hfftn()
    >>> print(B)
    [19.         -5.04891734 -0.30797853 -0.64310413 -0.64310413 -0.30797853
     -5.04891734]
    >>> assert id(B) == id(hfftn.output_array)
    >>> assert id(A) == id(hfftn.input_array)

    """
    kind = C2R
    assert input_array.dtype.char in 'FDG'
    sz = list(input_array.shape)
    if s is not None:
        assert len(axes) == len(s)
        for q, axis in zip(s, axes):
            sz[axis] = q
    else:
        sz[axes[-1]] = 2*sz[axes[-1]]-2
    if output_array is None:
        dtype = input_array.dtype.char
        n = get_alignment(input_array)
        output_array = aligned(sz, n=n, dtype=np.dtype(dtype.lower()))
    else:
        assert list(output_array.shape) == sz
    assert sz[axes[-1]]//2+1 == input_array.shape[axes[-1]]
    M = get_normalization(kind, sz, axes)
    return get_planned_FFT(input_array, output_array, axes, kind, threads,
                           flags, M)

def get_normalization(kind, shape, axes):
    """Return normalization factor for multidimensional transform

    The normalization factor is, for Fourier transforms::

        1./np.prod(np.take(shape, axes))

    where shape is the global shape of the array that is input to the
    forward transform, and axes are the axes transformed over.

    For real-to-real transforms the normalization factor for each axis is

        - REDFT00 - 2(N-1)
        - REDFT01 - 2N
        - REDFT10 - 2N
        - REDFT11 - 2N
        - RODFT00 - 2(N+1)
        - RODFT01 - 2N
        - RODFT10 - 2N
        - RODFT11 - 2N

    where N is the length of the input array along that axis.

    Parameters
    ----------
    kind : sequence of ints
        The kind of transform along each axis
    shape : sequence of ints
        The shape of the global transformed array (input to the forward
        transform)
    axes : sequence of ints
        The axes transformed over

    Note
    ----
    The returned normalization factor is the *inverse* of the product of the
    normalization factors for the axes it is transformed over.

    """
    kind = [kind]*len(axes) if isinstance(kind, int) else kind
    assert len(kind) == len(axes)
    M = 1
    for knd, axis in zip(kind, axes):
        N = shape[axis]
        if knd == FFTW_RODFT00:
            M *= 2*(N+1)
        elif knd == FFTW_REDFT00:
            M *= 2*(N-1)
        elif knd in (FFTW_RODFT01, FFTW_RODFT10, FFTW_RODFT11,
                     FFTW_REDFT01, FFTW_REDFT10, FFTW_REDFT11):
            M *= 2*N
        else:
            M *= N
    return 1./M

inverse = {
    FFTW_RODFT11: FFTW_RODFT11,
    FFTW_REDFT11: FFTW_REDFT11,
    FFTW_RODFT01: FFTW_RODFT10,
    FFTW_RODFT10: FFTW_RODFT01,
    FFTW_REDFT01: FFTW_REDFT10,
    FFTW_REDFT10: FFTW_REDFT01,
    FFTW_RODFT00: FFTW_RODFT00,
    FFTW_REDFT00: FFTW_REDFT00,
    rfftn: irfftn,
    irfftn: rfftn,
    fftn: ifftn,
    ifftn: fftn,
    dctn: idctn,
    idctn: dctn,
    dstn: idstn,
    idstn: dstn,
    hfftn: ihfftn,
    ihfftn: hfftn
}



import functools
import numpy as np
from . import fftw

def _Xfftn_plan_pyfftw(shape, axes, dtype, transforms, options):

    import pyfftw
    opts = dict(
        avoid_copy=True,
        overwrite_input=True,
        auto_align_input=True,
        auto_contiguous=True,
        threads=1,
    )
    opts.update(options)

    transforms = {} if transforms is None else transforms
    if tuple(axes) in transforms:
        plan_fwd, plan_bck = transforms[tuple(axes)]
    else:
        if np.issubdtype(dtype, np.floating):
            plan_fwd = pyfftw.builders.rfftn
            plan_bck = pyfftw.builders.irfftn
        else:
            plan_fwd = pyfftw.builders.fftn
            plan_bck = pyfftw.builders.ifftn

    s = tuple(np.take(shape, axes))
    U = pyfftw.empty_aligned(shape, dtype=dtype)
    xfftn_fwd = plan_fwd(U, s=s, axes=axes, **opts)
    U.fill(0)
    if np.issubdtype(dtype, np.floating):
        del opts['overwrite_input']
    V = xfftn_fwd.output_array
    xfftn_bck = plan_bck(V, s=s, axes=axes, **opts)
    V.fill(0)
    xfftn_fwd.update_arrays(U, V)
    xfftn_bck.update_arrays(V, U)

    wrapped_xfftn_bck = functools.partial(xfftn_bck, normalise_idft=False)
    functools.update_wrapper(wrapped_xfftn_bck, xfftn_bck,
                             assigned=['input_array',
                                       'output_array',
                                       '__doc__'])

    return (xfftn_fwd, wrapped_xfftn_bck)

def _Xfftn_plan_fftw(shape, axes, dtype, transforms, options):

    opts = dict(
        overwrite_input='FFTW_DESTROY_INPUT',
        planner_effort='FFTW_MEASURE',
        threads=1,
    )
    opts.update(options)
    flags = (fftw.flag_dict[opts['planner_effort']],
             fftw.flag_dict[opts['overwrite_input']])
    threads = opts['threads']

    transforms = {} if transforms is None else transforms
    if tuple(axes) in transforms:
        plan_fwd, plan_bck = transforms[tuple(axes)]
    else:
        if np.issubdtype(dtype, np.floating):
            plan_fwd = fftw.rfftn
            plan_bck = fftw.irfftn
        else:
            plan_fwd = fftw.fftn
            plan_bck = fftw.ifftn

    s = tuple(np.take(shape, axes))
    U = fftw.aligned(shape, dtype=dtype)
    xfftn_fwd = plan_fwd(U, s=s, axes=axes, threads=threads, flags=flags)
    U.fill(0)
    V = xfftn_fwd.output_array
    if np.issubdtype(dtype, np.floating):
        flags = (fftw.flag_dict[opts['planner_effort']],)
    xfftn_bck = plan_bck(V, s=s, axes=axes, threads=threads, flags=flags, output_array=U)
    return (xfftn_fwd, xfftn_bck)

def _Xfftn_plan_numpy(shape, axes, dtype, transforms, options):

    transforms = {} if transforms is None else transforms
    if tuple(axes) in transforms:
        plan_fwd, plan_bck = transforms[tuple(axes)]
    else:
        if np.issubdtype(dtype, np.floating):
            plan_fwd = np.fft.rfftn
            plan_bck = np.fft.irfftn
        else:
            plan_fwd = np.fft.fftn
            plan_bck = np.fft.ifftn

    s = tuple(np.take(shape, axes))
    U = fftw.aligned(shape, dtype=dtype)
    V = plan_fwd(U, s=s, axes=axes).astype(dtype.char.upper()) # Numpy returns complex double if input single precision
    V = fftw.aligned_like(V)
    M = np.prod(s)

    # Numpy has forward transform unscaled and backward scaled with 1/N
    return (_Yfftn_wrap(plan_fwd, U, V, 1, {'s': s, 'axes': axes}),
            _Yfftn_wrap(plan_bck, V, U, M, {'s': s, 'axes': axes}))

def _Xfftn_plan_mkl(shape, axes, dtype, transforms, options): #pragma: no cover

    transforms = {} if transforms is None else transforms
    if tuple(axes) in transforms:
        plan_fwd, plan_bck = transforms[tuple(axes)]
    else:
        if np.issubdtype(dtype, np.floating):
            from mkl_fft._numpy_fft import rfftn, irfftn
            plan_fwd = rfftn
            plan_bck = irfftn
        else:
            from mkl_fft._numpy_fft import fftn, ifftn
            plan_fwd = fftn
            plan_bck = ifftn

    s = tuple(np.take(shape, axes))
    U = fftw.aligned(shape, dtype=dtype)
    V = plan_fwd(U, s=s, axes=axes)
    V = fftw.aligned_like(V)
    M = np.prod(s)

    return (_Yfftn_wrap(plan_fwd, U, V, 1, {'s': s, 'axes': axes}),
            _Yfftn_wrap(plan_bck, V, U, M, {'s': s, 'axes': axes}))

def _Xfftn_plan_scipy(shape, axes, dtype, transforms, options):

    transforms = {} if transforms is None else transforms
    if tuple(axes) in transforms:
        plan_fwd, plan_bck = transforms[tuple(axes)]
    else:
        from scipy.fftpack import fftn, ifftn # No rfftn/irfftn methods
        plan_fwd = fftn
        plan_bck = ifftn

    s = tuple(np.take(shape, axes))
    U = fftw.aligned(shape, dtype=dtype)
    V = plan_fwd(U, shape=s, axes=axes)
    V = fftw.aligned_like(V)
    M = np.prod(s)
    return (_Yfftn_wrap(plan_fwd, U, V, 1, {'shape': s, 'axes': axes}),
            _Yfftn_wrap(plan_bck, V, U, M, {'shape': s, 'axes': axes}))

class _Yfftn_wrap(object):
    #Wraps numpy/scipy/mkl transforms to FFTW style
    # pylint: disable=too-few-public-methods

    __slots__ = ('_xfftn', '_M', '_opt', '__doc__', '_input_array', '_output_array')

    def __init__(self, xfftn_obj, input_array, output_array, M, opt):
        object.__setattr__(self, '_xfftn', xfftn_obj)
        object.__setattr__(self, '_opt', opt)
        object.__setattr__(self, '_M', M)
        object.__setattr__(self, '_input_array', input_array)
        object.__setattr__(self, '_output_array', output_array)
        object.__setattr__(self, '__doc__', xfftn_obj.__doc__)

    @property
    def input_array(self):
        return object.__getattribute__(self, '_input_array')

    @property
    def output_array(self):
        return object.__getattribute__(self, '_output_array')

    @property
    def xfftn(self):
        return object.__getattribute__(self, '_xfftn')

    @property
    def opt(self):
        return object.__getattribute__(self, '_opt')

    @property
    def M(self):
        return object.__getattribute__(self, '_M')

    def __call__(self, *args, **kwargs):
        self.opt.update(kwargs)
        self.output_array[...] = self.xfftn(self.input_array, **self.opt)
        if abs(self.M-1) > 1e-8:
            self._output_array *= self.M
        return self.output_array

class _Xfftn_wrap(object):
    #Common interface for all serial transforms
    # pylint: disable=too-few-public-methods

    __slots__ = ('_xfftn', '__doc__', '_input_array', '_output_array')

    def __init__(self, xfftn_obj, input_array, output_array):
        object.__setattr__(self, '_xfftn', xfftn_obj)
        object.__setattr__(self, '_input_array', input_array)
        object.__setattr__(self, '_output_array', output_array)
        object.__setattr__(self, '__doc__', xfftn_obj.__doc__)

    @property
    def input_array(self):
        return object.__getattribute__(self, '_input_array')

    @property
    def output_array(self):
        return object.__getattribute__(self, '_output_array')

    @property
    def xfftn(self):
        return object.__getattribute__(self, '_xfftn')

    def __call__(self, input_array=None, output_array=None, **options):
        if input_array is not None:
            self.input_array[...] = input_array
        self.xfftn(**options)
        if output_array is not None:
            output_array[...] = self.output_array
            return output_array
        else:
            return self.output_array

class FFTBase(object):
    """Base class for serial FFT transforms

    Parameters
    ----------
    shape : list or tuple of ints
        shape of input array planned for
    axes : None, int or tuple of ints, optional
        axes to transform over. If None transform over all axes
    dtype : np.dtype, optional
        Type of input array
    padding : bool, number or list of numbers
        If False, then no padding. If number, then apply this number as padding
        factor for all axes. If list of numbers, then each number gives the
        padding for each axis. Must be same length as axes.
    """

    def __init__(self, shape, axes=None, dtype=float, padding=False):
        shape = list(shape) if np.ndim(shape) else [shape]
        assert len(shape) > 0
        assert min(shape) > 0
        if axes is not None:
            axes = list(axes) if np.ndim(axes) else [axes]
            for i, axis in enumerate(axes):
                if axis < 0:
                    axes[i] = axis + len(shape)
        else:
            axes = list(range(len(shape)))
        assert min(axes) >= 0
        assert max(axes) < len(shape)
        assert 0 < len(axes) <= len(shape)
        assert sorted(axes) == sorted(set(axes))

        dtype = np.dtype(dtype)
        assert dtype.char in 'fdgFDG'
        self.shape = shape
        self.axes = axes
        self.dtype = dtype
        self.padding = padding
        self.real_transform = np.issubdtype(dtype, np.floating)
        self.padding_factor = 1

    def _truncation_forward(self, padded_array, trunc_array):
        axis = self.axes[-1]
        if self.padding_factor > 1.0+1e-8:
            trunc_array.fill(0)
            N0 = self.forward.output_array.shape[axis]
            if self.real_transform:
                N = trunc_array.shape[axis]
                s = [slice(None)]*trunc_array.ndim
                s[axis] = slice(0, N)
                trunc_array[:] = padded_array[tuple(s)]
                if N0 % 2 == 0:
                    s[axis] = N-1
                    s = tuple(s)
                    trunc_array[s] = trunc_array[s].real
                    trunc_array[s] *= 2
            else:
                N = trunc_array.shape[axis]
                su = [slice(None)]*trunc_array.ndim
                su[axis] = slice(0, N//2+1)
                trunc_array[tuple(su)] = padded_array[tuple(su)]
                su[axis] = slice(-(N//2), None)
                trunc_array[tuple(su)] += padded_array[tuple(su)]

    def _padding_backward(self, trunc_array, padded_array):
        axis = self.axes[-1]
        if self.padding_factor > 1.0+1e-8:
            padded_array.fill(0)
            N0 = self.forward.output_array.shape[axis]
            if self.real_transform:
                s = [slice(0, n) for n in trunc_array.shape]
                padded_array[tuple(s)] = trunc_array[:]
                N = trunc_array.shape[axis]
                if N0 % 2 == 0: # Symmetric Fourier interpolator
                    s[axis] = N-1
                    s = tuple(s)
                    padded_array[s] = padded_array[s].real
                    padded_array[s] *= 0.5
            else:
                N = trunc_array.shape[axis]
                su = [slice(None)]*trunc_array.ndim
                su[axis] = slice(0, N//2+1)
                padded_array[tuple(su)] = trunc_array[tuple(su)]
                su[axis] = slice(-(N//2), None)
                padded_array[tuple(su)] = trunc_array[tuple(su)]
                if N0 % 2 == 0:  # Use symmetric Fourier interpolator
                    su[axis] = N//2
                    padded_array[tuple(su)] *= 0.5
                    su[axis] = -(N//2)
                    padded_array[tuple(su)] *= 0.5


class FFT(FFTBase):
    """Class for serial FFT transforms

    Parameters
    ----------
    shape : list or tuple of ints
        shape of input array planned for
    axes : None, int or tuple of ints, optional
        axes to transform over. If None transform over all axes
    dtype : np.dtype, optional
        Type of input array
    padding : bool, number or list of numbers
        If False, then no padding. If number, then apply this number as padding
        factor for all axes. If list of numbers, then each number gives the
        padding for each axis. Must be same length as axes.
    backend : str, optional
        Choose backend for serial transforms (``fftw``, ``pyfftw``, ``numpy``,
        ``scipy``, ``mkl_fft``). Default is ``fftw``
    transforms : None or dict, optional
        Dictionary of axes to serial transforms (forward and backward) along
        those axes. For example::

            {(0, 1): (dctn, idctn), (2, 3): (dstn, idstn)}

        If missing the default is to use rfftn/irfftn for real input arrays and
        fftn/ifftn for complex input arrays. Real-to-real transforms can be
        configured using this dictionary and real-to-real transforms from the
        :mod:`.fftw.xfftn` module.
    kw : dict
        Parameters passed to serial transform object

    Methods
    -------
    forward(input_array=None, output_array=None, **options)
        Generic serial forward transform

        Parameters
        ----------
        input_array : array, optional
        output_array : array, optional
        options : dict
            parameters to serial transforms

        Returns
        -------
        output_array : array

    backward(input_array=None, output_array=None, **options)
        Generic serial backward transform

        Parameters
        ----------
        input_array : array, optional
        output_array : array, optional
        options : dict
            parameters to serial transforms

        Returns
        -------
        output_array : array

    """
    def __init__(self, shape, axes=None, dtype=float, padding=False,
                 backend='fftw', transforms=None, **kw):
        FFTBase.__init__(self, shape, axes, dtype, padding)
        plan = {
            'pyfftw': _Xfftn_plan_pyfftw,
            'fftw': _Xfftn_plan_fftw,
            'numpy': _Xfftn_plan_numpy,
            'mkl_fft': _Xfftn_plan_mkl,
            'scipy': _Xfftn_plan_scipy,
        }[backend]
        self.backend = backend
        self.fwd, self.bck = plan(self.shape, self.axes, self.dtype, transforms, kw)
        U, V = self.fwd.input_array, self.fwd.output_array
        self.M = 1
        if backend != 'fftw':
            self.M = 1./np.prod(np.take(self.shape, self.axes))
        else:
            self.M = self.fwd.get_normalization()
        if backend == 'scipy':
            self.real_transform = False # No rfftn/irfftn methods
        self.padding_factor = 1.0
        if padding is not False:
            self.padding_factor = padding[self.axes[-1]] if np.ndim(padding) else padding
        if abs(self.padding_factor-1.0) > 1e-8:
            assert len(self.axes) == 1
            trunc_array = self._get_truncarray(shape, V.dtype)
            self.forward = _Xfftn_wrap(self._forward, U, trunc_array)
            self.backward = _Xfftn_wrap(self._backward, trunc_array, U)
        else:
            self.forward = _Xfftn_wrap(self._forward, U, V)
            self.backward = _Xfftn_wrap(self._backward, V, U)

    def _forward(self, **kw):
        normalize = kw.pop('normalize', True)
        self.fwd(None, None, **kw)
        self._truncation_forward(self.fwd.output_array, self.forward.output_array)
        if normalize:
            self.forward._output_array *= self.M
        return self.forward.output_array

    def _backward(self, **kw):
        normalize = kw.pop('normalize', False)
        self._padding_backward(self.backward.input_array, self.bck.input_array)
        self.bck(None, None, **kw)
        if normalize:
            self.backward._output_array *= self.M
        return self.backward.output_array

    def _get_truncarray(self, shape, dtype):
        axis = self.axes[-1]
        if not self.real_transform:
            shape = list(shape)
            shape[axis] = int(np.round(shape[axis] / self.padding_factor))
            return fftw.aligned(shape, dtype=dtype)

        shape = list(shape)
        shape[axis] = int(np.round(shape[axis] / self.padding_factor))
        shape[axis] = shape[axis]//2 + 1
        return fftw.aligned(shape, dtype=dtype)



from mpi4py import MPI
import numpy as np

__all__ = ('FileBase',)

comm = MPI.COMM_WORLD

class FileBase(object):
    """Base class for reading/writing distributed arrays

    Parameters
    ----------
    filename : str, optional
        Name of backend file used to store data
    domain : sequence, optional
        An optional spatial mesh or domain to go with the data.
        Sequence of either

            - 2-tuples, where each 2-tuple contains the (origin, length)
              of each dimension, e.g., (0, 2*pi).
            - Arrays of coordinates, e.g., np.linspace(0, 2*pi, N). One
              array per dimension.

    """
    def __init__(self, filename=None, domain=None):
        self.f = None
        self.filename = filename
        self.domain = domain

    def _check_domain(self, group, field):
        """Check dimensions and store (if missing) self.domain"""
        raise NotImplementedError

    def write(self, step, fields, **kw):
        """Write snapshot ``step`` of ``fields`` to file

        Parameters
        ----------
        step : int
            Index of snapshot.
        fields : dict
            The fields to be dumped to file. (key, value) pairs are group name
            and either arrays or 2-tuples, respectively. The arrays are complete
            arrays to be stored, whereas 2-tuples are arrays with associated
            *global* slices.
        as_scalar : boolean, optional
            Whether to store rank > 0 arrays as scalars. Default is False.
        """
        as_scalar = kw.get("as_scalar", False)

        def _write(group, u, sl, step, kw, k=None):
            if sl is None:
                self._write_group(group, u, step, **kw)
            else:
                self._write_slice_step(group, step, sl, u, **kw)

        for group, list_of_fields in fields.items():
            assert isinstance(list_of_fields, (tuple, list))
            assert isinstance(group, str)

            for field in list_of_fields:
                u = field[0] if isinstance(field, (tuple, list)) else field
                sl = field[1] if isinstance(field, (tuple, list)) else None
                if as_scalar is False or u.rank == 0:
                    self._check_domain(group, u)
                    _write(group, u, sl, step, kw)
                else: # as_scalar is True and u.rank > 0
                    if u.rank == 1:
                        for k in range(u.shape[0]):
                            g = group + str(k)
                            self._check_domain(g, u[k])
                            _write(g, u[k], sl, step, kw)
                    elif u.rank == 2:
                        for k in range(u.shape[0]):
                            for l in range(u.shape[1]):
                                g = group + str(k) + str(l)
                                self._check_domain(g, u[k, l])
                                _write(g, u[k, l], sl, step, kw)

    def read(self, u, name, **kw):
        """Read field ``name`` into distributed array ``u``

        Parameters
        ----------
        u : array
            The :class:`.DistArray` to read into.
        name : str
            Name of field to be read.
        step : int, optional
            Index of field to be read. Default is 0.
        """
        raise NotImplementedError

    def close(self):
        """Close the self.filename file"""
        self.f.close()

    def open(self, mode='r+'):
        """Open the self.filename file for reading or writing

        Parameters
        ----------
        mode : str
           Open file in this mode. Default is 'r+'.
        """
        raise NotImplementedError

    @staticmethod
    def backend():
        """Return which backend is used to store data"""
        raise NotImplementedError

    def _write_slice_step(self, name, step, slices, field, **kwargs):
        raise NotImplementedError

    def _write_group(self, name, u, step, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _get_slice_name(slices):
        sl = list(slices)
        slname = ''
        for ss in sl:
            if isinstance(ss, slice):
                slname += 'slice_'
            else:
                slname += str(ss)+'_'
        return slname[:-1]

    @staticmethod
    def _get_local_slices(slices, s):
        # Check if data is on this processor and make slices local
        inside = 1
        si = np.nonzero([isinstance(x, int) and not z == slice(None) for x, z in zip(slices, s)])[0]
        for i in si:
            if slices[i] >= s[i].start and slices[i] < s[i].stop:
                slices[i] -= s[i].start
            else:
                inside = 0
        return slices, inside



import numpy as np
from mpi4py import MPI
from .file_base import FileBase

__all__ = ('HDF5File',)

comm = MPI.COMM_WORLD

class HDF5File(FileBase):
    """Class for reading/writing data to HDF5 format

    Parameters
    ----------
    h5name : str
        Name of hdf5 file to be created.
    domain : sequence, optional
        An optional spatial mesh or domain to go with the data.
        Sequence of either

            - 2-tuples, where each 2-tuple contains the (origin, length)
              of each dimension, e.g., (0, 2*pi).
            - Arrays of coordinates, e.g., np.linspace(0, 2*pi, N). One
              array per dimension.
    mode : str, optional
        ``r``, ``w`` or ``a`` for read, write or append. Default is ``a``.
    kw : dict, optional
        Optional additional keyword arguments used when creating the file
        used to store data.
    """
    def __init__(self, h5name, domain=None, mode='a', **kw):
        FileBase.__init__(self, h5name, domain=domain)
        import h5py
        self.f = h5py.File(h5name, mode, driver="mpio", comm=comm, **kw)
        self.close()

    def _check_domain(self, group, field):
        if self.domain is None:
            self.domain = ((0, 2*np.pi),)*field.dimensions
        assert len(self.domain) == field.dimensions
        self.f.require_group(group)
        if not "shape" in self.f[group].attrs:
            self.f[group].attrs.create("shape", field.pencil.shape)
        if not "rank" in self.f[group].attrs:
            self.f[group].attrs.create("rank", field.rank)
        assert field.rank == self.f[group].attrs["rank"]
        assert np.all(field.pencil.shape == self.f[group].attrs["shape"])
        if isinstance(self.domain[0], np.ndarray):
            self.f[group].require_group("mesh")
        else:
            self.f[group].require_group("domain")
        for i in range(field.dimensions):
            d = self.domain[i]
            if isinstance(d, np.ndarray):
                d0 = np.squeeze(d)
                self.f[group]["mesh"].require_dataset("x{}".format(i),
                                                      shape=d0.shape,
                                                      dtype=d0.dtype,
                                                      data=d0)
            else:
                d0 = np.array([d[0], d[1]])
                self.f[group]["domain"].require_dataset("x{}".format(i),
                                                        shape=d0.shape,
                                                        dtype=d0.dtype,
                                                        data=d0)

    @staticmethod
    def backend():
        return 'hdf5'

    def open(self, mode='r+'):
        import h5py
        self.f = h5py.File(self.filename, mode, driver="mpio", comm=comm)

    def write(self, step, fields, **kw):
        """Write snapshot ``step`` of ``fields`` to HDF5 file

        Parameters
        ----------
        step : int
            Index of snapshot.
        fields : dict
            The fields to be dumped to file. (key, value) pairs are group name
            and either arrays or 2-tuples, respectively. The arrays are complete
            arrays to be stored, whereas 2-tuples are arrays with associated
            *global* slices.
        as_scalar : boolean, optional
            Whether to store rank > 0 arrays as scalars. Default is False.

        Example
        -------
        >>> from mpi4py import MPI
        >>> from mpi4py_fft import PFFT, HDF5File, newDistArray
        >>> comm = MPI.COMM_WORLD
        >>> T = PFFT(comm, (15, 16, 17))
        >>> u = newDistArray(T, forward_output=False, val=1)
        >>> v = newDistArray(T, forward_output=False, val=2)
        >>> f = HDF5File('h5filename.h5', mode='w')
        >>> f.write(0, {'u': [u, (u, [slice(None), 4, slice(None)])],
        ...             'v': [v, (v, [slice(None), 5, 5])]})
        >>> f.write(1, {'u': [u, (u, [slice(None), 4, slice(None)])],
        ...             'v': [v, (v, [slice(None), 5, 5])]})

        This stores data within two main groups ``u`` and ``v``. The HDF5 file
        will in the end contain groups::

            /u/3D/{0, 1}
            /u/2D/slice_4_slice/{0, 1}
            /v/3D/{0, 1}
            /v/1D/slice_5_5/{0, 1}

        Note
        ----
        The list of slices used in storing only parts of the arrays are views
        of the *global* arrays.

        """
        self.open()
        FileBase.write(self, step, fields, **kw)
        self.close()

    def read(self, u, name, **kw):
        step = kw.get('step', 0)
        self.open()
        s = u.local_slice()
        dset = "/".join((name, "{}D".format(u.dimensions), str(step)))
        u[:] = self.f[dset][s]
        self.close()

    def _write_slice_step(self, name, step, slices, field, **kw):
        rank = field.rank
        slices = (slice(None),)*rank + tuple(slices)
        slices = list(slices)
        ndims = slices[rank:].count(slice(None))
        slname = self._get_slice_name(slices[rank:])
        s = field.local_slice()
        slices, inside = self._get_local_slices(slices, s)
        sp = np.nonzero([isinstance(x, slice) for x in slices])[0]
        sf = tuple(np.take(s, sp))
        sl = tuple(slices)
        group = "/".join((name, "{}D".format(ndims), slname))
        self.f.require_group(group)
        N = field.global_shape
        self.f[group].require_dataset(str(step), shape=tuple(np.take(N, sp)), dtype=field.dtype)
        if inside == 1:
            self.f["/".join((group, str(step)))][sf] = field[sl]

    def _write_group(self, name, u, step, **kw):
        s = u.local_slice()
        group = "/".join((name, "{}D".format(u.dimensions)))
        self.f.require_group(group)
        self.f[group].require_dataset(str(step), shape=u.global_shape, dtype=u.dtype)
        self.f["/".join((group, str(step)))][s] = u



cdef extern from "fftw3.h":

    ctypedef struct fftw_complex_struct:
        pass

    ctypedef fftw_complex_struct *fftw_complex

    ctypedef struct fftw_plan_struct:
        pass

    ctypedef fftw_plan_struct *fftw_plan

    void fftw_destroy_plan(fftw_plan)

    void fftw_execute_dft(fftw_plan, void *_in, void *_out) nogil

    void fftw_execute_dft_c2r(fftw_plan, void *_in, void *_out) nogil

    void fftw_execute_dft_r2c(fftw_plan, void *_in, void *_out) nogil

    void fftw_execute_r2r(fftw_plan, void *_in, void *_out) nogil

    void fftw_execute(fftw_plan) nogil

    void fftw_init_threads()

    void fftw_plan_with_nthreads(int n)

    int fftw_export_wisdom_to_filename(const char *filename)

    int fftw_import_wisdom_from_filename(const char *filename)

    void fftw_forget_wisdom()

    void fftw_set_timelimit(double seconds)

    void fftw_cleanup()

    void fftw_cleanup_threads()

    int fftw_alignment_of(void *_in)

    void fftw_print_plan(fftw_plan)


cdef extern from "fftw_planxfftn.h":

    ctypedef double fftw_real

    fftw_plan fftw_planxfftn(int      ndims,
                             int      sizes_in[],
                             void     *_in,
                             int      sizes_out[],
                             void     *_out,
                             int      naxes,
                             int      axes[],
                             int      kind[],
                             unsigned flags)

ctypedef void (*generic_function)(void *plan, void *_in, void *_out) nogil



"""
This is the **mpi4py-fft** package

What is **mpi4py-fft**?
=======================

The Python package **mpi4py-fft** is a tool primarily for working with Fast
Fourier Transforms (FFTs) of (large) multidimensional arrays. There is really
no limit as to how large the arrays can be, just as long as there is sufficient
computing powers available. Also, there are no limits as to how transforms can
be configured. Just about any combination of transforms from the FFTW library
is supported. Furthermore, **mpi4py-fft** can also be used simply to perform
global redistributions (distribute and communicate) of large arrays with MPI,
without any transforms at all.

For more information, see `documentation <https://mpi4py-fft.readthedocs.io>`_.

"""
__version__ = '2.0.5'
__author__ = 'Lisandro Dalcin and Mikael Mortensen'

from .distarray import DistArray, newDistArray, Function
from .mpifft import PFFT
from . import fftw
from .fftw import fftlib
from .io import HDF5File, NCFile, generate_xdmf
